package net.tamas.bcpg;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.openpgp.*;

import java.io.*;
import java.security.NoSuchProviderException;
import java.security.Security;
import java.security.SignatureException;
import java.util.Iterator;
import java.util.logging.Logger;

public class DecryptAndVerifyFile implements Runnable
{

  private Logger LOG = Logger.getAnonymousLogger();

  private static final String OPT_DECRYPT_RING_FILE = "d";
  private static final String OPT_DECRYPT_PASSPHRASE = "p";
  private static final String OPT_VERIFY_RING_FILE = "v";
  private static final String OPT_INPUT_FILE = "i";
  private static final String OPT_OUTPUT_FILE = "o";


  private CommandLine cl;

  private PGPPublicKeyRing verifyRing;
  private PGPSecretKeyRing decryptRing;
  private PGPPrivateKey decryptPrivateKey;

  public DecryptAndVerifyFile(CommandLine cl)
  {
    this.cl = cl;
  }

  public static void main(String[] args) throws Exception
  {
    Options options = new Options();
    options.addOption(OPT_DECRYPT_RING_FILE, true, "private keyring for decryption");
    options.addOption(OPT_DECRYPT_PASSPHRASE, true, "decryption keyring passphrase");
    options.addOption(OPT_VERIFY_RING_FILE, true, "public keyring for verification");
    options.addOption(OPT_INPUT_FILE, true, "input file name");
    options.addOption(OPT_OUTPUT_FILE, true, "output file name");
    CommandLineParser parser = new PosixParser();
    CommandLine cl = parser.parse(options, args);
    new DecryptAndVerifyFile(cl).run();
  }

  public void run()
  {
    setupBc();
    try
    {
      readKeys();
      findDecryptKey();
      decrypt();
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  private void readKeys() throws IOException
  {
    verifyRing = (PGPPublicKeyRing) readFirstObject(cl.getOptionValue(OPT_VERIFY_RING_FILE));
    decryptRing = (PGPSecretKeyRing) readFirstObject(cl.getOptionValue(OPT_DECRYPT_RING_FILE));
  }

  private void setupBc()
  {
    Security.addProvider(new BouncyCastleProvider());
    PGPUtil.setDefaultProvider("BC");
  }

  private void findDecryptKey() throws NoSuchProviderException, PGPException
  {
    // assumes only one key in the ring...
    Iterator i = decryptRing.getSecretKeys();
    PGPSecretKey decryptPgpSecretKey = (PGPSecretKey) i.next();
    char[] pass = cl.getOptionValue(OPT_DECRYPT_PASSPHRASE).toCharArray();
    decryptPrivateKey = decryptPgpSecretKey.extractPrivateKey(pass, "BC");
  }

  private void decrypt() throws IOException, NoSuchProviderException, PGPException, SignatureException
  {
    // read encrypted data from the file
    FileBasedPGPObjectFactory encryptedDataFactory = new FileBasedPGPObjectFactory(cl.getOptionValue(OPT_INPUT_FILE));
    PGPEncryptedDataList encryptedDataList = (PGPEncryptedDataList) encryptedDataFactory.nextObject();
    PGPPublicKeyEncryptedData encryptedData = (PGPPublicKeyEncryptedData) encryptedDataList.get(0);
    InputStream decryptedIn = encryptedData.getDataStream(decryptPrivateKey, "BC");
    // decompress the decrypted data stream
    PGPObjectFactory compressedFactory = new PGPObjectFactory(decryptedIn);
    PGPCompressedData compressed = (PGPCompressedData) compressedFactory.nextObject();
    PGPObjectFactory dataFactory = new PGPObjectFactory(compressed.getDataStream());
    // process the actual signatures and literal data
    PGPOnePassSignatureList opsList = (PGPOnePassSignatureList) dataFactory.nextObject();
    PGPOnePassSignature ops = opsList.get(0);
    ops.initVerify(verifyRing.getPublicKey(), "BC");
    PGPLiteralData literalData = (PGPLiteralData) dataFactory.nextObject();
    LOG.info("processing file '" + literalData.getFileName() + "'");
    FileOutputStream out = new FileOutputStream(cl.getOptionValue(OPT_OUTPUT_FILE));
    try
    {
      InputStream literalIn = literalData.getDataStream();
      int ch;
      while ((ch = literalIn.read()) >= 0)
      {
        ops.update((byte) ch);
        out.write(ch);
      }
      PGPSignatureList signatureList = (PGPSignatureList) dataFactory.nextObject();
      PGPSignature signature = signatureList.get(0);
      if (!ops.verify(signature))
        throw new SignatureException("Verify failed");

    }
    finally
    {
      out.flush();
      out.close();
    }
  }

  // also determines whether key files are ascii-armored
  private Object readFirstObject(String fileName) throws IOException
  {
    FileBasedPGPObjectFactory factory = new FileBasedPGPObjectFactory(fileName);
    Object result = factory.nextObject();
    factory.close();
    return result;
  }

}